package be.iminds.ilabt.jfed.lowlevel.api.test;

import be.iminds.ilabt.jfed.fedmon.webapi.service.json.Server;
import be.iminds.ilabt.jfed.fedmon.webapi.service.json.ServerBuilder;
import be.iminds.ilabt.jfed.fedmon.webapi.service.json.ServiceBuilder;
import be.iminds.ilabt.jfed.log.Logger;
import be.iminds.ilabt.jfed.lowlevel.api.RestApi;
import be.iminds.ilabt.jfed.lowlevel.connection.HttpConnection;
import be.iminds.ilabt.jfed.lowlevel.connection.JFedConnection;
import be.iminds.ilabt.jfed.lowlevel.connection.JFedException;
import be.iminds.ilabt.jfed.lowlevel.connection_pool.JFedConnectionProvider;
import be.iminds.ilabt.jfed.lowlevel.lib.*;
import be.iminds.ilabt.jfed.lowlevel.testbed_info.ApiInfo;
import be.iminds.ilabt.jfed.lowlevel.testbed_info.TestbedInfoSource;
import be.iminds.ilabt.jfed.lowlevel.user.GeniUserProvider;
import be.iminds.ilabt.jfed.preferences.JFedPreferences;
import be.iminds.ilabt.jfed.testing.base.*;
import be.iminds.ilabt.jfed.lowlevel.authority.legacy.TargetAuthority;
import com.fasterxml.jackson.annotation.JsonIgnoreProperties;
import com.fasterxml.jackson.annotation.JsonInclude;
import com.fasterxml.jackson.annotation.JsonProperty;
import com.fasterxml.jackson.core.type.TypeReference;
import com.fasterxml.jackson.databind.ObjectMapper;
import io.dropwizard.jackson.Jackson;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import javax.inject.Inject;
import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.*;
import java.util.stream.Collectors;

/**
 * A Test of GPULab
 */
public class TestGPULab extends BaseApiTest<GPULabTestConfig> {
    @Inject
    public TestGPULab(@Nonnull Logger logger,
                      @Nullable TargetAuthority testedAuthority,
                      @Nonnull GeniUserProvider geniUserProvider,
                      @Nonnull GPULabTestConfig testConfig,
                      @Nonnull JFedConnectionProvider connectionProvider,
                      @Nonnull TestbedInfoSource testbedInfoSource,
                      @Nonnull JFedPreferences jFedPreferences) {
        super(logger, testedAuthority, geniUserProvider, testConfig,
                connectionProvider, testbedInfoSource,
                jFedPreferences);
    }


    private static final ObjectMapper MAPPER = Jackson.newObjectMapper();
    private static final ApiTestMetaData metadata = new ApiTestMetaData<GPULabTestConfig>() {
        @Nonnull
        @Override
        public String getTestDescription() {
            return "GPULab Test: tests connectivity and basic functionality";
        }

        @Nonnull
        @Override
        public Class<GPULabTestConfig> getConfigClass() {
            return GPULabTestConfig.class;
        }

        @Nonnull
        @Override
        public GPULabTestConfig parseApiTestConfig(@Nonnull String apiTestConfigString) {
            if (apiTestConfigString == null || apiTestConfigString.trim().isEmpty())
                return new GPULabTestConfig();
            try {
                return MAPPER.readValue(apiTestConfigString, GPULabTestConfig.class);
            } catch (IOException e) {
                throw new RuntimeException("Invalid Test config", e);
            }
        }

        @Nonnull
        @Override
        public ApiTestConfigEditor<GPULabTestConfig> getApiTestConfigEditor() {
            return new JsonTestConfigEditor(GPULabTestConfig.class);
        }

    };
    public static ApiTestMetaData getMetaData() {
        return metadata;
    }


    private URL url;
    public void setUp() throws MalformedURLException {
        url = new URL(testConfig.getGpulabURL());

        note("done");
    }


    @Test(groups = {"test"})
    public void testGPULab() throws JFedException {
        GPULabApi api = new GPULabApi(logger, getJFedPreferences());
        ServiceBuilder serviceBuilder = new ServiceBuilder()
                .setApi(api.getApi().getName().getId())
                .setApiVersion(api.getApi().getVersion())
                .setUrl(testConfig.getGpulabURL())
                .setId(Integer.MAX_VALUE-5); //dummy ID
        Server server = new ServerBuilder()
                .setName("GPULab")
                .setBaseUrl(testConfig.getGpulabURL())
                .addServiceBuilder(serviceBuilder)
                .setId(Integer.MAX_VALUE-6) //dummy ID
                .create();
        JFedConnection con = connectionProvider.getConnectionByAuthority(user, server, api.getApi());

        List<GPULabCluster> clusters = api.getClusters((HttpConnection) con);
        assertNotEmpty(clusters, "There are no active GPULab clusters");
        note("GPULab version: "+clusters.get(0).getVersion());
        String clusterIds = clusters.stream().map(GPULabCluster::getId).collect(Collectors.joining(", "));
        note("Clusters: "+clusterIds);
        int allGpuCount = clusters.stream().map(GPULabCluster::getGpuTotalCount).filter(Objects::nonNull).mapToInt(Integer::intValue).sum();
        int freeGpuCount = clusters.stream().map(GPULabCluster::getGpuAvailableCount).filter(Objects::nonNull).mapToInt(Integer::intValue).sum();
        note("Total GPUs: "+allGpuCount);
        note("Free  GPUs: "+freeGpuCount);

        List<GPULabJob> jobs = api.getJobs((HttpConnection) con);
        note("Job count: "+jobs.size());
        note("Jobs: "+jobs.stream().map(GPULabJob::getUuid).collect(Collectors.joining(", ")));

        addExtraResult(new ApiTestResult.ApiTestMethodResult.FedmonResultExtraBuilder()
                .setKey("clusters")
                .setLarge(false)
                .setHttpMediaTypeToText()
                .setValue(clusters.size())
        );
        addExtraResult(new ApiTestResult.ApiTestMethodResult.FedmonResultExtraBuilder()
                .setKey("cluster_ids")
                .setLarge(false)
                .setHttpMediaTypeToText()
                .setValue(clusterIds)
        );
        addExtraResult(new ApiTestResult.ApiTestMethodResult.FedmonResultExtraBuilder()
                .setKey("recent_job_count")
                .setLarge(false)
                .setHttpMediaTypeToText()
                .setValue(jobs.size())
        );
        addExtraResult(new ApiTestResult.ApiTestMethodResult.FedmonResultExtraBuilder()
                .setKey("gpu")
                .setLarge(false)
                .setHttpMediaTypeToText()
                .setValue(freeGpuCount+"/"+allGpuCount)
        );

        note("Test completed successfully.");
    }



    public static class GPULabApi extends RestApi {
        public GPULabApi(@Nonnull Logger logger, @Nonnull JFedPreferences jFedPreferences) {
            super(logger, new ApiInfo.Api(ApiInfo.ApiName.GENERIC_HTTPS_CLIENTAUTH, 1), jFedPreferences);
        }

        @Nonnull
        @Override
        public String getName() {
            return getApiName();
        }

        /**
         * A human readable name for the implemented API
         *
         * @return "SLA Collector";
         */
        static public String getApiName() {
            return "GPULabApi";
        }

        @ApiMethod(order = 1, hint = "GET")
        public List<GPULabCluster> getClusters(@Nonnull HttpConnection con) throws JFedException {
            String javaName = "getClusters";
            String relUrl = "gpulab/api/v2.0/clusters";
            Map<String, Object> urlParams = makeMethodParameters();

            RestApi.RestReply<String> r = internalGet(con, javaName, relUrl, urlParams, urlParams);
            try {
                return MAPPER.readValue(r.getValue(), new TypeReference<List<GPULabCluster>>() {});
            } catch (IOException e) {
                throw new JFedException("Exception mapping reply", e);
            }
        }

        @ApiMethod(order = 1, hint = "GET")
        public List<GPULabJob> getJobs(@Nonnull HttpConnection con) throws JFedException {
            String javaName = "getJobs";
            String relUrl = "gpulab/api/v2.0/jobs?max_hours=900&max_count=100";
            Map<String, Object> urlParams = makeMethodParameters();

            RestApi.RestReply<String> r = internalGet(con, javaName, relUrl, urlParams, urlParams);
            try {
                return MAPPER.readValue(r.getValue(), new TypeReference<List<GPULabJob>>() {});
            } catch (IOException e) {
                throw new JFedException("Exception mapping reply", e);
            }
        }
    }

    @JsonInclude(JsonInclude.Include.NON_NULL)
    @JsonIgnoreProperties(ignoreUnknown=true)
    private static class GPULabCluster {
        @Nullable private final String id;
        @Nullable private final String comment;
        @Nullable private final Integer cpuAvailableCount;
        @Nullable private final Integer cpuTotalCount;
        @Nullable private final Integer gpuAvailableCount;
        @Nullable private final Integer gpuTotalCount;
        @Nullable private final List<String> gpuTypes;
        @Nullable private final Integer slaveCount;
        @Nullable private final String version;

        public GPULabCluster(@Nullable @JsonProperty("cluster_id") String id,
                             @Nullable @JsonProperty("comment") String comment,
                             @Nullable @JsonProperty("cpu_available_count") Integer cpuAvailableCount,
                             @Nullable @JsonProperty("cpu_total_count") Integer cpuTotalCount,
                             @Nullable @JsonProperty("gpu_available_count") Integer gpuAvailableCount,
                             @Nullable @JsonProperty("gpu_total_count") Integer gpuTotalCount,
                             @Nullable @JsonProperty("gpu_types") List<String> gpuTypes,
                             @Nullable @JsonProperty("slave_count") Integer slaveCount,
                             @Nullable @JsonProperty("version") String version) {
            this.id = id;
            this.comment = comment;
            this.cpuAvailableCount = cpuAvailableCount;
            this.cpuTotalCount = cpuTotalCount;
            this.gpuAvailableCount = gpuAvailableCount;
            this.gpuTotalCount = gpuTotalCount;
            this.gpuTypes = gpuTypes;
            this.slaveCount = slaveCount;
            this.version = version;
        }

        @Nullable
        @JsonProperty("cluster_id")
        public String getId() {
            return id;
        }

        @Nullable
        @JsonProperty("comment")
        public String getComment() {
            return comment;
        }

        @JsonProperty("cpu_available_count")
        @Nullable
        public Integer getCpuAvailableCount() {
            return cpuAvailableCount;
        }

        @JsonProperty("cpu_total_count")
        @Nullable
        public Integer getCpuTotalCount() {
            return cpuTotalCount;
        }

        @JsonProperty("gpu_available_count")
        @Nullable
        public Integer getGpuAvailableCount() {
            return gpuAvailableCount;
        }

        @JsonProperty("gpu_total_count")
        @Nullable
        public Integer getGpuTotalCount() {
            return gpuTotalCount;
        }

        @JsonProperty("gpu_types")
        @Nullable
        public List<String> getGpuTypes() {
            return gpuTypes;
        }

        @JsonProperty("slave_count")
        @Nullable
        public Integer getSlaveCount() {
            return slaveCount;
        }

        @JsonProperty("version")
        @Nullable
        public String getVersion() {
            return version;
        }

        @Override
        public String toString() {
            return "GPULabCluster{" +
                    "id='" + id + '\'' +
                    ", comment='" + comment + '\'' +
                    ", cpuAvailableCount='" + cpuAvailableCount + '\'' +
                    ", cpuTotalCount='" + cpuTotalCount + '\'' +
                    ", gpuAvailableCount='" + gpuAvailableCount + '\'' +
                    ", gpuTotalCount='" + gpuTotalCount + '\'' +
                    ", gpuTypes=" + gpuTypes +
                    ", slaveCount=" + slaveCount +
                    ", version='" + version + '\'' +
                    '}';
        }
    }

    @JsonInclude(JsonInclude.Include.NON_NULL)
    @JsonIgnoreProperties(ignoreUnknown=true)
    private static class GPULabJob {
        @Nullable private final String uuid;
        @Nullable private final String created;
        @Nullable private final String version;
        @Nullable private final String name;
//        @Nullable private final String nvdocker_data;
        @Nullable private final String project;
        @Nullable private final String queueTime;
        @Nullable private final String state;
        @Nullable private final String username;
//        @Nullable private final Map run_details; //deadlineData endDate slaveDnsName slaveHostname startDate


        public GPULabJob(@JsonProperty("uuid") @Nullable String uuid,
                         @JsonProperty("created") @Nullable String created,
                         @JsonProperty("version") @Nullable String version,
                         @JsonProperty("name") @Nullable String name,
                         @JsonProperty("project") @Nullable String project,
                         @JsonProperty("queue_time") @Nullable String queueTime,
                         @JsonProperty("state") @Nullable String state,
                         @JsonProperty("username") @Nullable String username) {
            this.uuid = uuid;
            this.created = created;
            this.version = version;
            this.name = name;
            this.project = project;
            this.queueTime = queueTime;
            this.state = state;
            this.username = username;
        }

        @JsonProperty
        @Nullable
        public String getUuid() {
            return uuid;
        }

        @JsonProperty
        @Nullable
        public String getCreated() {
            return created;
        }

        @JsonProperty
        @Nullable
        public String getVersion() {
            return version;
        }

        @JsonProperty
        @Nullable
        public String getName() {
            return name;
        }

        @JsonProperty
        @Nullable
        public String getProject() {
            return project;
        }

        @JsonProperty("queue_time")
        @Nullable
        public String getQueueTime() {
            return queueTime;
        }

        @JsonProperty
        @Nullable
        public String getState() {
            return state;
        }

        @JsonProperty
        @Nullable
        public String getUsername() {
            return username;
        }

        @Override
        public String toString() {
            return "GPULabJob{" +
                    "uuid='" + uuid + '\'' +
                    ", created='" + created + '\'' +
                    ", version='" + version + '\'' +
                    ", name='" + name + '\'' +
                    ", project='" + project + '\'' +
                    ", queueTime='" + queueTime + '\'' +
                    ", state='" + state + '\'' +
                    ", username='" + username + '\'' +
                    '}';
        }
    }
}
