package be.iminds.ilabt.jfed.experimenter_gui.editor;

import be.iminds.ilabt.jfed.fedmon.webapi.service.json.Resource;
import be.iminds.ilabt.jfed.fedmon.webapi.service.json.ResourceClass;
import be.iminds.ilabt.jfed.fedmon.webapi.service.json.Server;
import be.iminds.ilabt.jfed.fedmon.webapi.service.json.Testbed;
import be.iminds.ilabt.util.jsonld.impl.PrimaryIdObject;
import com.fasterxml.jackson.annotation.*;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import java.util.*;
import java.util.function.Predicate;

/**
 * A toolbox filter.
 *
 * All filters are ** OR'ed **
 */
@JsonInclude(JsonInclude.Include.NON_EMPTY)
@JsonIgnoreProperties(ignoreUnknown=true)
public class ToolBoxFilter {
    @Nonnull private final SortedSet<String> testbedIds;
    @Nonnull private final SortedSet<String> organisationIds;
    @Nonnull private final SortedSet<String> categoryNames;
    @Nonnull private final SortedSet<String> federationIds;

    /** Should intra-category filters be OR? (If no, AND is used) */
    private final boolean orIntra;

    /**
     * Default ToolBoxFilter matches everything
     */
    public ToolBoxFilter() {
        this.testbedIds = Collections.emptySortedSet();
        this.organisationIds = Collections.emptySortedSet();
        this.categoryNames = Collections.emptySortedSet();
        this.federationIds = Collections.emptySortedSet();
        this.orIntra = false;
    }

    @JsonCreator
    public ToolBoxFilter(@Nullable @JsonProperty("testbedIds") Collection<String> testbedIds,
                         @Nullable @JsonProperty("organisationIds") Collection<String> organisationIds,
                         @Nullable @JsonProperty("categoryNames") Collection<String> categoryNames,
                         @Nullable @JsonProperty("federationIds") Collection<String> federationIds,
                         @Nullable @JsonProperty("orIntra") Boolean orIntra) {
        this.testbedIds = testbedIds == null ?
                Collections.emptySortedSet() :
                Collections.unmodifiableSortedSet(new TreeSet<>(testbedIds));
        this.organisationIds = organisationIds == null ?
                Collections.emptySortedSet() :
                Collections.unmodifiableSortedSet(new TreeSet<>(organisationIds));
        this.categoryNames = categoryNames == null ?
                Collections.emptySortedSet() :
                Collections.unmodifiableSortedSet(new TreeSet<>(categoryNames));
        this.federationIds = federationIds == null ?
                Collections.emptySortedSet() :
                Collections.unmodifiableSortedSet(new TreeSet<>(federationIds));
        this.orIntra = Objects.equals(orIntra, Boolean.TRUE);
    }

    @Nonnull
    @JsonProperty
    public SortedSet<String> getTestbedIds() {
        return testbedIds;
    }

    @Nonnull
    @JsonProperty
    public SortedSet<String> getOrganisationIds() {
        return organisationIds;
    }

    @Nonnull
    @JsonProperty
    public SortedSet<String> getFederationIds() {
        return federationIds;
    }

    @Nonnull
    @JsonProperty
    public SortedSet<String> getCategoryNames() {
        return categoryNames;
    }

    @JsonProperty
    public boolean isOrIntra() {
        return orIntra;
    }

    /**
     * If the filter is empty, it matches everything
     * */
    @JsonIgnore
    public boolean isEmpty() {
        return testbedIds.isEmpty() && organisationIds.isEmpty() && categoryNames.isEmpty() && federationIds.isEmpty();
    }


    @JsonIgnore
    public boolean serverMatches(@Nullable Server server) {
        if (server == null)
            return false;

        if (server.getTestbedId() == null || server.getTestbed() == null)
            return false;
        Testbed testbed = server.getTestbed();
        assert Objects.equals(testbed.getId(), server.getTestbedId());
        String testbedId = server.getTestbedId();

        boolean testbedMatch = testbedIds.stream().anyMatch(testbedId::equals);
        if (testbedMatch && orIntra) return true;
        if (!testbedIds.isEmpty() && !testbedMatch && !orIntra) return false;

        if (testbed.getFederations() != null) {
            boolean anyFedMatch = false;
            for (String federationId : federationIds) {
                if (testbed.getFederations().stream()
                        .map(PrimaryIdObject::getId)
                        .anyMatch(federationId::equals))
                    anyFedMatch = true;
            }
            if (anyFedMatch && orIntra) return true;
            if (!federationIds.isEmpty() && !anyFedMatch && !orIntra) return false;
        }
        if (testbed.getOrganisation() != null) {
            boolean anyOrgMatch = false;
            if (organisationIds.stream()
                    .anyMatch((oid) -> testbed.getOrganisation().getId().equals(oid)))
                anyOrgMatch = true;
            if (anyOrgMatch && orIntra) return true;
            if (!organisationIds.isEmpty() && !anyOrgMatch && !orIntra) return false;
        }
        if (testbed.getCategoryNames() != null) {
            boolean anyCatMatch = false;
            if (!Collections.disjoint(testbed.getCategoryNames(), categoryNames))
                anyCatMatch = true;
            if (anyCatMatch && orIntra) return true;
            if (!categoryNames.isEmpty() && !anyCatMatch && !orIntra) return false;
        }

        return !orIntra;
    }

    @Nonnull
    @JsonIgnore
    public Predicate<Server> getServerFilter() {
        //if no filter, all match
        if (isEmpty())
            return (s) -> true;

        return this::serverMatches;
    }

    @Nonnull
    @JsonIgnore
    public Predicate<Resource> getResourceFilter() {
        //if no filter, all match
        if (isEmpty())
            return (r) -> true;

        return (resource) -> serverMatches(resource.getServer());
    }

    @Nonnull
    @JsonIgnore
    public Predicate<ResourceClass> getResourceClassFilter() {
        //if no filter, all match
        if (isEmpty())
            return (rc) -> true;

        return (resourceClass) -> {
            //don't show generic node
            if (!isEmpty() &&
                    resourceClass.getId().toLowerCase().startsWith("generic-node"))
                return false;

            //show all other matching classes
            return resourceClass.getResources()
                    .stream()
                    .anyMatch(getResourceFilter());
        };
    }

    @Override
    public boolean equals(Object o) {
        if (this == o) return true;
        if (o == null || getClass() != o.getClass()) return false;
        ToolBoxFilter that = (ToolBoxFilter) o;
        return Objects.equals(testbedIds, that.testbedIds) &&
                Objects.equals(organisationIds, that.organisationIds) &&
                Objects.equals(categoryNames, that.categoryNames) &&
                Objects.equals(federationIds, that.federationIds) &&
                orIntra == ((ToolBoxFilter) o).orIntra;
    }

    @Override
    public int hashCode() {
        return Objects.hash(testbedIds, organisationIds, categoryNames, federationIds, orIntra);
    }
}
